from gc import collect, get_referrers
from typing import Any

from pytest import MonkeyPatch


def _is_referrer_a_module(d: dict) -> bool:
    return isinstance(d, dict) and "__loader__" in d.keys()


def replace_bound_references_in_memory(
    monkeypatch_ctx: MonkeyPatch,
    target_obj: Any,
    new_obj: Any,
) -> None:
    collect()  # Perform GC before checking for references in memory
    for reference in get_referrers(target_obj):
        if _is_referrer_a_module(reference):
            for k, v in reference.items():
                if v is target_obj:
                    monkeypatch_ctx.setitem(reference, k, new_obj)



from pybond.assertions import (
    called_exactly_once_with_args,
    called_with_args,
    called_with_exact_args_list,
    times_called,
    was_called,
)
from pybond.james import calls, spy, stub

__all__ = [
    "called_exactly_once_with_args",
    "called_with_args",
    "called_with_exact_args_list",
    "calls",
    "spy",
    "stub",
    "times_called",
    "was_called",
]



from types import ModuleType
from typing import Any, Callable, Tuple, TypeAlias, TypedDict

FunctionCall = TypedDict(
    "FunctionCall",
    {
        "args": list[Any],
        "kwargs": dict[str, Any],
        "error": Any,
        "return": Any,
    }
)

Spyable: TypeAlias = Callable | object
SpyTarget: TypeAlias = Tuple[ModuleType, str]
StubTarget: TypeAlias = Tuple[ModuleType, str, Spyable]



from inspect import getfullargspec
from typing import Callable


def _args_match(fsig, gsig) -> bool:
    if len(fsig.args) == 0 or len(gsig.args) == 0:
        return fsig.args == gsig.args
    else:
        return len(fsig.args) == len(gsig.args)


def _varargs_match(fsig, gsig) -> bool:
    if None in [fsig.varargs, gsig.varargs]:
        return fsig.varargs == gsig.varargs
    else:
        return True


def _kwargs_match(fsig, gsig) -> bool:
    if len(fsig.kwonlyargs) == 0 or len(gsig.kwonlyargs) == 0:
        return fsig.kwonlyargs == gsig.kwonlyargs
    else:
        fdefaults = {} if fsig.kwonlydefaults is None else fsig.kwonlydefaults
        gdefaults = {} if gsig.kwonlydefaults is None else gsig.kwonlydefaults
        return (
            set(fsig.kwonlyargs) == set(gsig.kwonlyargs) and
            set(fdefaults.keys()) == set(gdefaults.keys())
        )


def _varkwargs_match(fsig, gsig) -> bool:
    if None in [fsig.varkw, gsig.varkw]:
        return fsig.varkw == gsig.varkw
    else:
        return True


def _fn_with_zero_arguments():
    return None


def function_signatures_match(f, g):
    try:
        fsig = getfullargspec(f)
        gsig = getfullargspec(g)
        return (
            _args_match(fsig, gsig)
            and _kwargs_match(fsig, gsig)
            and _varargs_match(fsig, gsig)
            and _varkwargs_match(fsig, gsig)
        )
    except TypeError as e:
        # Some callables may not be introspectable in certain implementations of
        # Python. For example, in CPython, some built-in functions defined in C
        # provide no metadata about their arguments.
        if str(e) == "unsupported callable":
            if [f.__module__, f.__name__] == ["time", "time"]:
                return function_signatures_match(_fn_with_zero_arguments, g)
            # Add other specific cases here
            else:
                raise
        else:
            raise


def is_wrapped_function(f: Callable) -> bool:
    return hasattr(f, "__wrapped__")



"""This module is inspired by clojure's bond library."""

import sys
from contextlib import contextmanager
from copy import deepcopy
from functools import wraps
from inspect import isclass
from typing import Callable

from pytest import MonkeyPatch

from pybond.memory import replace_bound_references_in_memory
from pybond.util import function_signatures_match, is_wrapped_function
from pybond.types import FunctionCall, Spyable, SpyTarget, StubTarget


def _function_call(args, kwargs, error, return_value) -> FunctionCall:
    return {
        "args": args,
        "kwargs": kwargs,
        "error": error,
        "return": return_value,
    }


def _spy_function(f: Callable) -> Spyable:
    """
    Wrap f, returning a new function that keeps track of its call count and
    arguments.
    """
    _calls = []

    def calls():
        return _calls

    @wraps(f)
    def handle_function_call(*args, **kwargs):
        # Assume the worst: f might mutate its arguments
        non_mutated_args = deepcopy(list(args)) if args else None
        non_mutated_kwargs = deepcopy(dict(kwargs)) if kwargs else None
        try:
            return_value = f(*args, **kwargs)
            _calls.append(
                _function_call(
                    args=non_mutated_args,
                    kwargs=non_mutated_kwargs,
                    error=None,
                    return_value=return_value,
                )
            )
            return return_value
        except Exception:
            _calls.append(
                _function_call(
                    args=non_mutated_args,
                    kwargs=non_mutated_kwargs,
                    error=sys.exc_info(),
                    return_value=None,
                )
            )
            raise

    handle_function_call.__wrapped__ = f
    setattr(handle_function_call, "calls", calls)
    return handle_function_call


def calls(f: Spyable) -> list[FunctionCall]:
    """
    Takes one arg, a function that has previously been spied. Returns a list of
    function call dicts, one per call. Each object contains the keys `args`,
    `kwargs`, `error` and `return_value`.

    If the function has not been spied, raises an exception.
    """
    if hasattr(f, "calls") and callable(f):
        return getattr(f, "calls")()
    else:
        raise ValueError(
            "The argument is not a spied function. Calls of an unspied "
            "function are not tracked and are therefore not known."
        )


def _function_signatures_match(originalf: Callable, stubf: Callable) -> bool:
    """
    Supports both regular functions and decorated functions using
    functools.wraps
    """
    return (
        (
            is_wrapped_function(originalf)
            and function_signatures_match(originalf.__wrapped__, stubf)
        ) or (
            not is_wrapped_function(originalf)
            and function_signatures_match(originalf, stubf)
        )
    )


def _check_if_class_is_instrumentable(
    original_obj: Spyable,
    stub_obj: Spyable,
    strict: bool = True,
) -> None:
    # TODO: implement spying on classes and class methods
    return


def _check_if_function_is_instrumentable(
    original_obj: Callable,
    stub_obj: Callable,
    strict: bool = True,
) -> None:
    if strict and not _function_signatures_match(original_obj, stub_obj):
        raise ValueError(
            f"Stub does not match the signature of {original_obj.__name__}."
        )


def _instrumented_obj(
    original_obj: Spyable,
    stub_obj: Spyable,
    strict: bool = True,
) -> Spyable:
    if isclass(original_obj):
        # TODO: implement spying on classes and class methods
        _check_if_class_is_instrumentable(original_obj, stub_obj, strict)
        return stub_obj
    elif callable(original_obj) and callable(stub_obj):
        _check_if_function_is_instrumentable(original_obj, stub_obj, strict)
        return _spy_function(stub_obj)
    elif callable(original_obj) and not callable(stub_obj):
        raise ValueError(
            f"Provided stub for Callable {original_obj.__name__} of type "
            f"{type(stub_obj)} is invalid: pybond expected a Callable type."
        )
    else:
        raise ValueError(
            f"Object of type {type(original_obj)} is not supported by pybond."
        )


@contextmanager
def stub(*targets: StubTarget, strict: bool = True):
    """
    Context manager which takes a list of targets to stub and spy on.

    Example usage:

    ```
    import my_module

    with stub([my_module, "test_function", lambda x: 42]):
        assert my_module.test_function("abc") == 42  # True
        function_calls = calls(my_module.test_function)
    ```
    """
    with MonkeyPatch.context() as m:
        try:
            for module, obj_name, stub_obj in targets:
                original_obj = getattr(module, obj_name)
                new_obj = _instrumented_obj(original_obj, stub_obj, strict)

                # The following only covers imports in the form:
                #     `import some_module`
                m.setattr(target=module, name=obj_name, value=new_obj)

                # The following covers bound imports in the form:
                #     `from some_module import some_object`
                replace_bound_references_in_memory(m, original_obj, new_obj)

            yield
        except Exception:
            raise


@contextmanager
def spy(*targets: SpyTarget):
    """
    Context manager which takes a list of targets to spy on.

    Example usage:

    ```
    import my_module
    with spy([my_module, "test_function"]):
        my_module.test_function("abc")
        function_calls = calls(my_module.test_function)
    ```
    """
    with stub(*[(m, n, deepcopy(getattr(m, n))) for m, n in targets]):
        yield



from pybond.james import calls


def was_called(f):
    """
    A predicate to check if `f` was called at least 1 time. Note that `f` must
    be a spied function.
    """
    return len(calls(f)) > 0


def times_called(f, n):
    """
    A predicate to check if `f` was called exactly `n` times. Note that `f` must
    be a spied function.
    """
    return len(calls(f)) == n


def called_with_exact_args_list(f, args_list=None, kwargs_list=None):
    """
    A predicate to check if `f` was called with specific arguments. Return true
    only if arguments match for every function call on `f`. Note that `f` must
    be a spied function.
    """
    fcalls = calls(f)
    if len(fcalls) == 0:
        return False
    args_match = len(fcalls) > 0
    kwargs_match = len(fcalls) > 0
    if args_list is not None:
        args_match = args_list == [fcall["args"] for fcall in fcalls]
    if kwargs_list is not None:
        kwargs_match = kwargs_list == [fcall["kwargs"] for fcall in fcalls]
    return args_match and kwargs_match


def called_exactly_once_with_args(f, args=None, kwargs=None):
    """
    A predicate to check if `f` was called with specific arguments exactly once.
    Note that `f` must be a spied function.
    """
    return called_with_exact_args_list(f, [args], [kwargs])


def called_with_args(f, args=None, kwargs=None):
    """
    A predicate to check if `f` has been called at least once with the given
    arguments. Note that `f` must be a spied function.
    """
    fcalls = calls(f)
    args_match = len(fcalls) > 0
    kwargs_match = len(fcalls) > 0
    if args is not None:
        args_match = args in [fcall["args"] for fcall in fcalls]
    if kwargs is not None:
        kwargs_match = kwargs in [fcall["kwargs"] for fcall in fcalls]
    return args_match and kwargs_match
